/* -----------------------------------------------------------------
 * SUNDIALS Copyright Start
 * Copyright (c) 2025-2026, Lawrence Livermore National Security,
 * University of Maryland Baltimore County, and the SUNDIALS contributors.
 * Copyright (c) 2013-2025, Lawrence Livermore National Security
 * and Southern Methodist University.
 * Copyright (c) 2002-2013, Lawrence Livermore National Security.
 * All rights reserved.
 *
 * See the top-level LICENSE and NOTICE files for details.
 *
 * SPDX-License-Identifier: BSD-3-Clause
 * SUNDIALS Copyright End
 * -----------------------------------------------------------------*/

#ifndef _SUNDIALS_CORE_HPP
#define _SUNDIALS_CORE_HPP

#include <sundials/sundials_context.hpp>
#include <sundials/sundials_core.h>
#include <sundials/sundials_linearsolver.hpp>
#include <sundials/sundials_matrix.hpp>
#include <sundials/sundials_memory.hpp>
#include <sundials/sundials_nonlinearsolver.hpp>
#include <sundials/sundials_nvector.hpp>
#include <sundials/sundials_profiler.hpp>

#endif /* _SUNDIALS_CORE_HPP */
